/****************************************************************************************
*  tcluucod.c ;  A tcl loadable moduale to provide uuencode and uudecode
*                functions for tcl.  These functions are simply a modified
*                version of the GNU uucode.c and uuencode.c and are subject
*                to the GNU license agreement.  See GNU.lic for details.
*  Having these as tcl functions allows me to create a tcl script that will
*  run under dos and UNIX.
***************************************************************************************/
#include <tcl.h>


#include <stdio.h>
#include <ctype.h>
#include <string.h>

#ifndef Windows
#include <pwd.h>
#else
#include <io.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#ifndef S_IRWXG
#define S_IRWXG S_IRWXU
#endif

#ifndef S_IRWXO
#define S_IRWXO S_IRWXU
#endif



static int	uuencode _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));
static int	uudecode _ANSI_ARGS_((ClientData clientData,
			 Tcl_Interp *interp, int argc, char **argv));

#define HAVE_FCHMODE


/* ENC is the basic 1 character encoding function to make a char printing.  */
#define	ENC(c) ((c) ? ((c) & 077) + ' ': '`')
#define	DEC(c) (((c) - ' ') & 077)

#if S_IRWXU != 0700
#error
choke me - Must translate mode argument (For uuencode )
#endif
static int uuencode(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{
  register int n, ch;
  register char *p;
  char buf[80],*uuname, *decname;
	FILE *infile,*outfile;
	struct stat sb;
	int mode;

	if ( argc < 3 || argc > 4)
		{
		sprintf(interp->result,"#args %s file [decname] encname",argv[0]);
		return TCL_ERROR;
		}
	if( argc == 3 ) 
		{
		decname = argv[1]; /* Same as original */
		uuname  = argv[2];
		}
	else
		{
		decname = argv[2]; /* Same as original */
		uuname  = argv[3];
		}

	if ( (infile = fopen(argv[1],"rb")) == NULL)
		{
		sprintf(interp->result,"Can't open %s for input (%d)",argv[1],errno);
		return TCL_ERROR;
		}

	if ( (outfile = fopen(uuname,"w")) == NULL)
		{
		sprintf(interp->result,"Can't open %s for output (%d)",argv[2],errno);
		fclose(infile);
		return TCL_ERROR;
		}

      if (fstat (fileno (infile), &sb))
		{
		sprintf(interp->result,"Can,t stat %s (%d)",argv[1],errno);
		fclose(infile);
		fclose(outfile);
		return TCL_ERROR;
		}

      mode = sb.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO);

  fprintf (outfile,"begin %o %s\n", mode, decname);

  while ((n = fread (buf, 1, 45, infile)) != 0)
    {
      ch = ENC (n);
      if (putc((char)ch,outfile) == EOF)
	break;
      for (p = buf; n > 0; n -= 3, p += 3)
	{
	  ch = *p >> 2;
	  ch = ENC (ch);
	  if (putc((char)ch,outfile) == EOF)
	    break;
	  ch = ((*p << 4) & 060) | ((p[1] >> 4) & 017);
	  ch = ENC (ch);
	  if (putc((char)ch,outfile) == EOF)
	    break;
	  ch = ((p[1] << 2) & 074) | ((p[2] >> 6) & 03);
	  ch = ENC (ch);
	  if (putc((char)ch,outfile) == EOF)
	    break;
	  ch = p[2] & 077;
	  ch = ENC (ch);
	  if (putc((char)ch,outfile) == EOF)
	    break;
	}
      if (putc('\n',outfile) == EOF)
	break;
    }

  if (ferror (infile))
    	{
	sprintf(interp->result, "read error");
  	fclose(infile);
  	fclose(outfile);
	return TCL_ERROR;
	}
  ch = ENC ('\0');
  putc((char)ch,outfile);
  putc('\n',outfile);

  fprintf (outfile,"end\n");
  if (ferror (outfile))
    	{
	sprintf(interp->result, "write error");
  	fclose(infile);
  	fclose(outfile);
	return TCL_ERROR;
	}
  fclose(infile);
  fclose(outfile);
  interp->result = "";
  return TCL_OK;
}



/********************************************************************************************
*  uudecode filename
*         filename = the file to be uudecoded
*******************************************************************************************/
static char retlst [1024];
static int uudecode(dummy, interp, argc, argv)
	ClientData dummy;		/* Not used. */
	Tcl_Interp *interp;		/* Current interpreter. */
	int argc;			/* Number of arguments. */
	char **argv;			/* Argument strings. */
{

  register int n;
  register char ch, *p;
  int mode,count;
  char buf[2 * BUFSIZ];
  char *outname;
  FILE *infile,*outfile;

	if ( argc != 2 ) 
		{
		sprintf(interp->result , "Wrong # of args (%s filename)",argv[0]);
		return TCL_ERROR;
		}

  if( (infile = (fopen(argv[1],"r"))) == NULL)
	{
	sprintf(interp->result , "Can't open %s (%d)",argv[1],errno);
	return TCL_ERROR;
	}
  /* Search for header line.  */

  *retlst = '\0';
  count = 0;

retry :
  do
    {
      if (fgets (buf, sizeof (buf), infile) == NULL)
	{
  	fclose(infile);
  	if (!count)
		{
		  sprintf (interp->result, "%s: no `begin' line", argv[1]);
		  return TCL_ERROR;
		}
      	else
		{
		 /* This is the GOOD exit point */
		  sprintf (interp->result, "%s", retlst);
		  return TCL_OK;
		}
	}
    }
  while (strncmp (buf, "begin ", 6) != 0)
    ;

  sscanf (buf, "begin %o %s", &mode, buf);

  outname = buf;

  if ( count == 0 )
  	strcpy(retlst,outname);
  else
	{
  	strcat(retlst," ");
  	strcat(retlst,outname);
	}

  /* Create output file and set mode.  */

  if ((outfile=fopen (outname, "wb")) == NULL)
    {
      sprintf (interp->result, "%s: %s", outname, argv[1]);
      fclose(infile); 
      return TCL_ERROR;
    }

  
  /* If we can't change the mode, we'll create the file anyway */
#if HAVE_FCHMOD      
      fchmod (fileno (outfile), mode & (S_IRWXU | S_IRWXG | S_IRWXO));
#else
      chmod (outname, mode & (S_IRWXU | S_IRWXG | S_IRWXO));
#endif

  /* For each input line:  */

  while (1)
    {
      if (fgets (buf, sizeof(buf), infile) == NULL)
	{
	  sprintf (interp->result, "%s: short file", argv[1]);
      	  fclose(infile); 
      	  fclose(outfile); 
	  return TCL_ERROR;
	}
      p = buf;

      /* N is used to avoid writing out all the characters at the end of
	 the file.  */

      n = DEC (*p);
      if (n <= 0)
	break;
      for (++p; n > 0; p += 4, n -= 3)
	{
	  if (n >= 3)
	    {
	      ch = (DEC (p[0]) << 2 )|( DEC (p[1]) >> 4);
	      putc((char)ch,outfile);
	      ch = (DEC (p[1]) << 4 )|( DEC (p[2]) >> 2);
	      putc((char)ch,outfile);
	      ch = (DEC (p[2]) << 6 )|( DEC (p[3]));
	      putc((char)ch,outfile);
	    }
	  else
	    {
	      if (n >= 1)
		{
		  ch = (DEC (p[0]) << 2 ) | ( DEC (p[1]) >> 4);
		  putc((char)ch,outfile);
		}
	      if (n >= 2)
		{
		  ch = (DEC (p[1]) << 4) | (DEC (p[2]) >> 2) ;
		  putc((char)ch,outfile);
		}
	    }
	}
    }

  if (fgets (buf, sizeof(buf), infile) == NULL
      || strcmp (buf, "end\n"))
    {
      sprintf (interp->result, "%s: no `end' line", argv[1]);
      fclose(infile);
      fclose(outfile); 
      return TCL_ERROR;
    }
  fclose(outfile); 

  count++;
  goto retry;
  return TCL_OK;
}


/*
/*
 *----------------------------------------------------------------------
 *
 * Uucode_Init -- the library initialization file for the extension
 *
 */


int 
#ifdef Windows
_export
#endif 
Uucode_Init(interp)
	 Tcl_Interp *interp;		
{

	 Tcl_CreateCommand(interp, "uuencode", uuencode, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	 Tcl_CreateCommand(interp, "uudecode", uudecode, (ClientData) 0,
		 (Tcl_CmdDeleteProc *) NULL);
	return TCL_OK;
}


